import logging
import threading
import time

from typing import Callable, Dict, Optional, Iterator  # noqa

from chalice.cli.filewatch import FileWatcher, WorkerProcess
from chalice.utils import OSUtils


LOGGER = logging.getLogger(__name__)


class StatWorkerProcess(WorkerProcess):
    def _start_file_watcher(self, project_dir):
        # type: (str) -> None
        watcher = StatFileWatcher()
        watcher.watch_for_file_changes(project_dir, self._on_file_change)

    def _on_file_change(self):
        # type: () -> None
        self._restart_event.set()


class StatFileWatcher(FileWatcher):
    POLL_INTERVAL = 1

    def __init__(self, osutils=None):
        # type: (Optional[OSUtils]) -> None
        self._mtime_cache = {}  # type: Dict[str, float]
        self._shutdown_event = threading.Event()
        self._thread = None  # type: Optional[threading.Thread]
        if osutils is None:
            osutils = OSUtils()
        self._osutils = osutils

    def watch_for_file_changes(self, root_dir, callback):
        # type: (str, Callable[[], None]) -> None
        t = threading.Thread(target=self.poll_for_changes_until_shutdown,
                             args=(root_dir, callback))
        t.daemon = True
        t.start()
        self._thread = t
        LOGGER.debug("Stat file watching: %s, with callback: %s",
                     root_dir, callback)

    def poll_for_changes_until_shutdown(self, root_dir, callback):
        # type: (str, Callable[[], None]) -> None
        self._seed_mtime_cache(root_dir)
        while not self._shutdown_event.is_set():
            self._single_pass_poll(root_dir, callback)
            time.sleep(self.POLL_INTERVAL)

    def _seed_mtime_cache(self, root_dir):
        # type: (str) -> None
        for rootdir, _, filenames in self._osutils.walk(root_dir):
            for filename in filenames:
                path = self._osutils.joinpath(rootdir, filename)
                self._mtime_cache[path] = self._osutils.mtime(path)

    def _single_pass_poll(self, root_dir, callback):
        # type: (str, Callable[[], None]) -> None
        new_mtimes = {}  # type: Dict[str, float]
        for path in self._recursive_walk_files(root_dir):
            if self._is_changed_file(path, new_mtimes):
                callback()
                return
        if new_mtimes != self._mtime_cache:
            # Files were removed.
            LOGGER.debug("Files removed, triggering restart.")
            self._mtime_cache = new_mtimes
            callback()
            return

    def _is_changed_file(self, path, new_mtimes):
        # type: (str, Dict[str, float]) -> bool
        last_mtime = self._mtime_cache.get(path)
        if last_mtime is None:
            LOGGER.debug("File added: %s, triggering restart.", path)
            return True
        try:
            new_mtime = self._osutils.mtime(path)
            if new_mtime > last_mtime:
                LOGGER.debug("File updated: %s, triggering restart.", path)
                return True
            new_mtimes[path] = new_mtime
            return False
        except (OSError, IOError):
            return False

    def _recursive_walk_files(self, root_dir):
        # type: (str) -> Iterator[str]
        for rootdir, _, filenames in self._osutils.walk(root_dir):
            for filename in filenames:
                path = self._osutils.joinpath(rootdir, filename)
                yield path



import copy
import re

from typing import List, Dict, Any, Optional, Callable  # noqa
import yaml

from chalice.config import Config  # noqa
from chalice import constants
from chalice import __version__ as chalice_version


def create_buildspec_v2(pipeline_params):
    # type: (PipelineParameters) -> Dict[str, Any]
    install_commands = [
        "pip install 'chalice%s'" % pipeline_params.chalice_version_range,
        "pip install -r requirements.txt",
    ]
    build_commands = [
        "chalice package /tmp/packaged",
        ("aws cloudformation package --template-file /tmp/packaged/sam.json "
         "--s3-bucket ${APP_S3_BUCKET} "
         "--output-template-file transformed.yaml")
    ]
    buildspec = {
        "version": "0.2",
        "phases": {
            "install": {
                "commands": install_commands,
                "runtime-versions": {
                    "python": pipeline_params.py_major_minor,
                }
            },
            "build": {
                "commands": build_commands,
            }
        },
        "artifacts": {
            "type": "zip",
            "files": [
                "transformed.yaml"
            ]
        }

    }
    return buildspec


def create_buildspec_legacy(pipeline_params):
    # type: (PipelineParameters) -> Dict[str, Any]
    install_commands = [
        'sudo pip install --upgrade awscli',
        'aws --version',
        "sudo pip install 'chalice%s'" % pipeline_params.chalice_version_range,
        'sudo pip install -r requirements.txt',
        'chalice package /tmp/packaged',
        ('aws cloudformation package '
            '--template-file /tmp/packaged/sam.json'
            ' --s3-bucket ${APP_S3_BUCKET} '
            '--output-template-file transformed.yaml'),
    ]
    buildspec = {
        'version': '0.1',
        'phases': {
            'install': {
                'commands': install_commands,
            }
        },
        'artifacts': {
            'type': 'zip',
            'files': ['transformed.yaml']
        }
    }
    return buildspec


class InvalidCodeBuildPythonVersion(Exception):
    def __init__(self, version, msg=None):
        # type: (str, Optional[str]) -> None
        if msg is None:
            msg = 'CodeBuild does not yet support python version %s.' % version
        super(InvalidCodeBuildPythonVersion, self).__init__(msg)


class PipelineParameters(object):

    _PYTHON_VERSION = re.compile('python(.+)')

    def __init__(self, app_name, lambda_python_version,
                 codebuild_image=None, code_source='codecommit',
                 chalice_version_range=None, pipeline_version='v1'):
        # type: (str, str, Optional[str], str, Optional[str], str) -> None
        self.app_name = app_name
        # lambda_python_version is what matches lambda, e.g. 'python3.7'.
        self.lambda_python_version = lambda_python_version
        # py_major_minor is just the version string, e.g. '3.7'
        self.py_major_minor = self._extract_version(lambda_python_version)
        self.codebuild_image = codebuild_image
        self.code_source = code_source
        if chalice_version_range is None:
            chalice_version_range = self._lock_to_minor_version()
        self.chalice_version_range = chalice_version_range
        self.pipeline_version = pipeline_version

    def _extract_version(self, lambda_python_version):
        # type: (str) -> str
        matched = self._PYTHON_VERSION.match(lambda_python_version)
        if matched is None:
            raise InvalidCodeBuildPythonVersion(lambda_python_version)
        return matched.group(1)

    def _lock_to_minor_version(self):
        # type: () -> str
        parts = [int(p) for p in chalice_version.split('.')]
        min_version = '%s.%s.%s' % (parts[0], parts[1], 0)
        max_version = '%s.%s.%s' % (parts[0], parts[1] + 1, 0)
        return '>=%s,<%s' % (min_version, max_version)


class BasePipelineTemplate(object):
    def create_template(self, pipeline_params):
        # type: (PipelineParameters) -> Dict[str, Any]
        raise NotImplementedError("create_template")


class CreatePipelineTemplateV2(BasePipelineTemplate):
    _BASE_TEMPLATE = {
        "AWSTemplateFormatVersion": "2010-09-09",
        "Parameters": {
            "ApplicationName": {
                "Default": "ChaliceApp",
                "Type": "String",
                "Description": "Enter the name of your application"
            },
            "CodeBuildImage": {
                "Default": "aws/codebuild/amazonlinux2-x86_64-standard:3.0",
                "Type": "String",
                "Description": "Name of codebuild image to use."
            }
        },
        "Resources": {},
        "Outputs": {},
    }

    def create_template(self, pipeline_params):
        # type: (PipelineParameters) -> Dict[str, Any]
        self._validate_python_version(pipeline_params.py_major_minor)
        t = copy.deepcopy(self._BASE_TEMPLATE)  # type: Dict[str, Any]
        params = t['Parameters']
        params['ApplicationName']['Default'] = pipeline_params.app_name
        resources = []  # type: List[BaseResource]
        if pipeline_params.code_source == 'github':
            resources.append(GithubSource())
        else:
            resources.append(CodeCommitSourceRepository())
        resources.extend([CodeBuild(create_buildspec_v2), CodePipeline()])
        for resource in resources:
            resource.add_to_template(t, pipeline_params)
        return t

    def _validate_python_version(self, python_version):
        # type: (str) -> None
        major, minor = [
            int(v) for v in python_version.split('.')
        ]
        if (major, minor) < (3, 7):
            raise InvalidCodeBuildPythonVersion(
                python_version,
                'This CodeBuild image does not support python version: %s' % (
                    python_version
                )
            )


class CreatePipelineTemplateLegacy(BasePipelineTemplate):

    _CODEBUILD_IMAGE = {
        'python2.7': 'python:2.7.12',
        'python3.6': 'python:3.6.5',
        'python3.7': 'python:3.7.1',
    }

    _BASE_TEMPLATE = {
        "AWSTemplateFormatVersion": "2010-09-09",
        "Parameters": {
            "ApplicationName": {
                "Default": "ChaliceApp",
                "Type": "String",
                "Description": "Enter the name of your application"
            },
            "CodeBuildImage": {
                "Default": "aws/codebuild/python:2.7.12",
                "Type": "String",
                "Description": "Name of codebuild image to use."
            }
        },
        "Resources": {},
        "Outputs": {},
    }

    def create_template(self, pipeline_params):
        # type: (PipelineParameters) -> Dict[str, Any]
        t = copy.deepcopy(self._BASE_TEMPLATE)  # type: Dict[str, Any]
        params = t['Parameters']
        params['ApplicationName']['Default'] = pipeline_params.app_name
        params['CodeBuildImage']['Default'] = self._get_codebuild_image(
            pipeline_params)

        resources = []  # type: List[BaseResource]
        if pipeline_params.code_source == 'github':
            resources.append(GithubSource())
        else:
            resources.append(CodeCommitSourceRepository())
        resources.extend([CodeBuild(create_buildspec_legacy), CodePipeline()])
        for resource in resources:
            resource.add_to_template(t, pipeline_params)
        return t

    def _get_codebuild_image(self, params):
        # type: (PipelineParameters) -> str
        if params.codebuild_image is not None:
            return params.codebuild_image
        try:
            image_suffix = self._CODEBUILD_IMAGE[params.lambda_python_version]
            return 'aws/codebuild/%s' % image_suffix
        except KeyError as e:
            raise InvalidCodeBuildPythonVersion(str(e))


class BaseResource(object):
    def add_to_template(self, template, pipeline_params):
        # type: (Dict[str, Any], PipelineParameters) -> None
        raise NotImplementedError("add_to_template")


class CodeCommitSourceRepository(BaseResource):
    def add_to_template(self, template, pipeline_params):
        # type: (Dict[str, Any], PipelineParameters) -> None
        resources = template.setdefault('Resources', {})
        resources['SourceRepository'] = {
            "Type": "AWS::CodeCommit::Repository",
            "Properties": {
                "RepositoryName": {
                    "Ref": "ApplicationName"
                },
                "RepositoryDescription": {
                    "Fn::Sub": "Source code for ${ApplicationName}"
                }
            }
        }
        template.setdefault('Outputs', {})['SourceRepoURL'] = {
            "Value": {
                "Fn::GetAtt": "SourceRepository.CloneUrlHttp"
            }
        }


class GithubSource(BaseResource):
    def add_to_template(self, template, pipeline_params):
        # type: (Dict[str, Any], PipelineParameters) -> None
        # For the github source, we don't create a github repo,
        # we just wire it up in the code pipeline.  The
        # only thing we add to the template are parameters
        # we reference in other resources later.
        p = template.setdefault('Parameters', {})
        p['GithubOwner'] = {
            'Type': 'String',
            'Description': 'The github owner or org name of the repository.',
        }
        p['GithubRepoName'] = {
            'Type': 'String',
            'Description': 'The name of the github repository.',
        }
        if pipeline_params.pipeline_version == 'v1':
            p['GithubPersonalToken'] = {
                'Type': 'String',
                'Description': 'Personal access token for the github repo.',
                'NoEcho': True,
            }
        else:
            p['GithubRepoSecretId'] = {
                'Type': 'String',
                'Default': 'GithubRepoAccess',
                'Description': (
                    'The name/ID of the SecretsManager secret that '
                    'contains the personal access token for the github repo.'
                )
            }
            p['GithubRepoSecretJSONKey'] = {
                'Type': 'String',
                'Default': 'OAuthToken',
                'Description': (
                    'The name of the JSON key in the SecretsManager secret '
                    'that contains the personal access token for the '
                    'github repo.'
                )
            }


class CodeBuild(BaseResource):
    def __init__(self, buildspec_generator=create_buildspec_legacy):
        # type: (Callable[[PipelineParameters], Dict[str, Any]]) -> None
        self._buildspec_generator = buildspec_generator

    def add_to_template(self, template, pipeline_params):
        # type: (Dict[str, Any], PipelineParameters) -> None
        resources = template.setdefault('Resources', {})
        outputs = template.setdefault('Outputs', {})
        # Used to store the application source when the SAM
        # template is packaged.
        self._add_s3_bucket(resources, outputs)
        self._add_codebuild_role(resources, outputs)
        self._add_codebuild_policy(resources)
        self._add_package_build(resources, pipeline_params)

    def _add_package_build(self, resources, pipeline_params):
        # type: (Dict[str, Any], PipelineParameters) -> None
        resources['AppPackageBuild'] = {
            "Type": "AWS::CodeBuild::Project",
            "Properties": {
                "Artifacts": {
                    "Type": "CODEPIPELINE"
                },
                "Environment": {
                    "ComputeType": "BUILD_GENERAL1_SMALL",
                    "Image": {
                        "Ref": "CodeBuildImage"
                    },
                    "Type": "LINUX_CONTAINER",
                    "EnvironmentVariables": [
                        {
                            "Name": "APP_S3_BUCKET",
                            "Value": {
                                "Ref": "ApplicationBucket"
                            }
                        }
                    ]
                },
                "Name": {
                    "Fn::Sub": "${ApplicationName}Build"
                },
                "ServiceRole": {
                    "Fn::GetAtt": "CodeBuildRole.Arn"
                },
                "Source": {
                    "Type": "CODEPIPELINE",
                    "BuildSpec": yaml.dump(
                        self._buildspec_generator(pipeline_params),
                    ),
                }
            }
        }

    def _add_s3_bucket(self, resources, outputs):
        # type: (Dict[str, Any], Dict[str, Any]) -> None
        resources['ApplicationBucket'] = {'Type': 'AWS::S3::Bucket'}
        outputs['S3ApplicationBucket'] = {
            'Value': {'Ref': 'ApplicationBucket'}
        }

    def _add_codebuild_role(self, resources, outputs):
        # type: (Dict[str, Any], Dict[str, Any]) -> None
        resources['CodeBuildRole'] = {
            "Type": "AWS::IAM::Role",
            "Properties": {
                "AssumeRolePolicyDocument": {
                    "Version": "2012-10-17",
                    "Statement": [
                        {
                            "Action": [
                                "sts:AssumeRole"
                            ],
                            "Effect": "Allow",
                            "Principal": {
                                "Service": [
                                    {'Fn::Sub': 'codebuild.${AWS::URLSuffix}'}
                                ]
                            }
                        }
                    ]
                }
            }
        }
        outputs['CodeBuildRoleArn'] = {
            "Value": {
                "Fn::GetAtt": "CodeBuildRole.Arn"
            }
        }

    def _add_codebuild_policy(self, resources):
        # type: (Dict[str, Any]) -> None
        resources['CodeBuildPolicy'] = {
            "Type": "AWS::IAM::Policy",
            "Properties": {
                "PolicyName": "CodeBuildPolicy",
                "PolicyDocument": constants.CODEBUILD_POLICY,
                "Roles": [
                    {
                        "Ref": "CodeBuildRole"
                    }
                ]
            }
        }


class CodePipeline(BaseResource):
    def add_to_template(self, template, pipeline_params):
        # type: (Dict[str, Any], PipelineParameters) -> None
        resources = template.setdefault('Resources', {})
        outputs = template.setdefault('Outputs', {})
        self._add_pipeline(resources, pipeline_params)
        self._add_bucket_store(resources, outputs)
        self._add_codepipeline_role(resources, outputs)
        self._add_cfn_deploy_role(resources, outputs)

    def _add_cfn_deploy_role(self, resources, outputs):
        # type: (Dict[str, Any], Dict[str, Any]) -> None
        outputs['CFNDeployRoleArn'] = {
            'Value': {'Fn::GetAtt': 'CFNDeployRole.Arn'}
        }
        resources['CFNDeployRole'] = {
            'Type': 'AWS::IAM::Role',
            'Properties': {
                "Policies": [
                    {
                        "PolicyName": "DeployAccess",
                        "PolicyDocument": {
                            "Version": "2012-10-17",
                            "Statement": [
                                {
                                    "Action": "*",
                                    "Resource": "*",
                                    "Effect": "Allow"
                                }
                            ]
                        }
                    }
                ],
                "AssumeRolePolicyDocument": {
                    "Version": "2012-10-17",
                    "Statement": [
                        {
                            "Action": [
                                "sts:AssumeRole"
                            ],
                            "Effect": "Allow",
                            "Principal": {
                                "Service": [
                                    {'Fn::Sub':
                                     'cloudformation.${AWS::URLSuffix}'}
                                ]
                            }
                        }
                    ]
                }
            }
        }

    def _add_pipeline(self, resources, pipeline_params):
        # type: (Dict[str, Any], PipelineParameters) -> None
        properties = {
            'Name': {
                'Fn::Sub': '${ApplicationName}Pipeline'
            },
            'ArtifactStore': {
                'Type': 'S3',
                'Location': {'Ref': 'ArtifactBucketStore'},
            },
            'RoleArn': {
                'Fn::GetAtt': 'CodePipelineRole.Arn',
            },
            'Stages': self._create_pipeline_stages(pipeline_params),
        }
        resources['AppPipeline'] = {
            'Type': 'AWS::CodePipeline::Pipeline',
            'Properties': properties
        }

    def _create_pipeline_stages(self, pipeline_params):
        # type: (PipelineParameters) -> List[Dict[str, Any]]
        # The goal is to eventually allow a user to configure
        # the various stages they want created. For now, there's
        # a fixed list.
        stages = []
        source = self._create_source_stage(pipeline_params)
        if source:
            stages.append(source)
        stages.extend([self._create_build_stage(), self._create_beta_stage()])
        return stages

    def _code_commit_source(self):
        # type: () -> Dict[str, Any]
        return {
            "Name": "Source",
            "Actions": [
                {
                    "ActionTypeId": {
                        "Category": "Source",
                        "Owner": "AWS",
                        "Version": 1,
                        "Provider": "CodeCommit"
                    },
                    "Configuration": {
                        "BranchName": "master",
                        "RepositoryName": {
                            "Fn::GetAtt": "SourceRepository.Name"
                        }
                    },
                    "OutputArtifacts": [
                        {
                            "Name": "SourceRepo"
                        }
                    ],
                    "RunOrder": 1,
                    "Name": "Source"
                }
            ]
        }

    def _create_source_stage(self, pipeline_params):
        # type: (PipelineParameters) -> Dict[str, Any]
        if pipeline_params.code_source == 'codecommit':
            return self._code_commit_source()
        return self._github_source(pipeline_params.pipeline_version)

    def _github_source(self, pipeline_version):
        # type: (str) -> Dict[str, Any]
        oauth_token = {'Ref': 'GithubPersonalToken'}  # type: Dict[str, Any]
        if pipeline_version == 'v2':
            oauth_token = {
                "Fn::Join": [
                    "", ["{{resolve:secretsmanager:",
                         {"Ref": "GithubRepoSecretId"},
                         ":SecretString:",
                         {"Ref": "GithubRepoSecretJSONKey"},
                         "}}"]
                ]
            }
        return {
            'Name': 'Source',
            'Actions': [{
                "Name": "Source",
                "ActionTypeId": {
                    "Category": "Source",
                    "Owner": "ThirdParty",
                    "Version": "1",
                    "Provider": "GitHub"
                },
                'RunOrder': 1,
                'OutputArtifacts': [{
                    'Name': 'SourceRepo',
                }],
                'Configuration': {
                    'Owner': {'Ref': 'GithubOwner'},
                    'Repo': {'Ref': 'GithubRepoName'},
                    'OAuthToken': oauth_token,
                    'Branch': 'master',
                    'PollForSourceChanges': True,
                }
            }],
        }

    def _create_build_stage(self):
        # type: () -> Dict[str, Any]
        return {
            "Name": "Build",
            "Actions": [
                {
                    "InputArtifacts": [
                        {
                            "Name": "SourceRepo"
                        }
                    ],
                    "Name": "CodeBuild",
                    "ActionTypeId": {
                        "Category": "Build",
                        "Owner": "AWS",
                        "Version": "1",
                        "Provider": "CodeBuild"
                    },
                    "OutputArtifacts": [
                        {
                            "Name": "CompiledCFNTemplate"
                        }
                    ],
                    "Configuration": {
                        "ProjectName": {
                            "Ref": "AppPackageBuild"
                        }
                    },
                    "RunOrder": 1
                }
            ]
        }

    def _create_beta_stage(self):
        # type: () -> Dict[str, Any]
        return {
            "Name": "Beta",
            "Actions": [
                {
                    "ActionTypeId": {
                        "Category": "Deploy",
                        "Owner": "AWS",
                        "Version": "1",
                        "Provider": "CloudFormation"
                    },
                    "InputArtifacts": [
                        {
                            "Name": "CompiledCFNTemplate"
                        }
                    ],
                    "Name": "CreateBetaChangeSet",
                    "Configuration": {
                        "ActionMode": "CHANGE_SET_REPLACE",
                        "ChangeSetName": {
                            "Fn::Sub": "${ApplicationName}ChangeSet"
                        },
                        "RoleArn": {
                            "Fn::GetAtt": "CFNDeployRole.Arn"
                        },
                        "Capabilities": "CAPABILITY_IAM",
                        "StackName": {
                            "Fn::Sub": "${ApplicationName}BetaStack"
                        },
                        "TemplatePath": "CompiledCFNTemplate::transformed.yaml"
                    },
                    "RunOrder": 1
                },
                {
                    "RunOrder": 2,
                    "ActionTypeId": {
                        "Category": "Deploy",
                        "Owner": "AWS",
                        "Version": "1",
                        "Provider": "CloudFormation"
                    },
                    "Configuration": {
                        "StackName": {
                            "Fn::Sub": "${ApplicationName}BetaStack"
                        },
                        "ActionMode": "CHANGE_SET_EXECUTE",
                        "ChangeSetName": {
                            "Fn::Sub": "${ApplicationName}ChangeSet"
                        },
                        "OutputFileName": "StackOutputs.json"
                    },
                    "Name": "ExecuteChangeSet",
                    "OutputArtifacts": [
                        {
                            "Name": "AppDeploymentValues"
                        }
                    ]
                }
            ]
        }

    def _add_bucket_store(self, resources, outputs):
        # type: (Dict[str, Any], Dict[str, Any]) -> None
        resources['ArtifactBucketStore'] = {
            'Type': 'AWS::S3::Bucket',
            'Properties': {
                'VersioningConfiguration': {
                    'Status': 'Enabled'
                }
            }
        }
        outputs['S3PipelineBucket'] = {
            'Value': {'Ref': 'ArtifactBucketStore'}
        }

    def _add_codepipeline_role(self, resources, outputs):
        # type: (Dict[str, Any], Dict[str, Any]) -> None
        outputs['CodePipelineRoleArn'] = {
            'Value': {'Fn::GetAtt': 'CodePipelineRole.Arn'}
        }
        resources['CodePipelineRole'] = {
            "Type": "AWS::IAM::Role",
            "Properties": {
                "Policies": [
                    {
                        "PolicyName": "DefaultPolicy",
                        "PolicyDocument": constants.CODEPIPELINE_POLICY,
                    }
                ],
                "AssumeRolePolicyDocument": {
                    "Version": "2012-10-17",
                    "Statement": [
                        {
                            "Action": [
                                "sts:AssumeRole"
                            ],
                            "Effect": "Allow",
                            "Principal": {
                                "Service": [
                                    {'Fn::Sub': 'codepipeline'
                                                '.${AWS::URLSuffix}'}
                                ]
                            }
                        }
                    ]
                }
            }
        }


class BuildSpecExtractor(object):
    def extract_buildspec(self, template):
        # type: (Dict[str, Any]) -> str
        source = template['Resources']['AppPackageBuild'][
            'Properties']['Source']
        buildspec = source.pop('BuildSpec')
        return buildspec



import io
import os
import zipfile
import json
import contextlib
import tempfile
import re
import shutil
import sys
import tarfile
from datetime import datetime, timedelta
import subprocess
from os import PathLike  # noqa

from collections import OrderedDict # noqa
import click
from typing import IO, Dict, List, Any, Tuple, Iterator, BinaryIO, Text  # noqa
from typing import Optional, Union  # noqa
from typing import MutableMapping, Callable  # noqa
from typing import cast  # noqa
import dateutil.parser
from dateutil.tz import tzutc

from chalice.constants import WELCOME_PROMPT

OptInt = Optional[int]
OptBytes = Optional[bytes]
EnvVars = MutableMapping
StrPath = Union[str, 'PathLike[str]']


class AbortedError(Exception):
    pass


def to_cfn_resource_name(name):
    # type: (str) -> str
    """Transform a name to a valid cfn name.

    This will convert the provided name to a CamelCase name.
    It's possible that the conversion to a CFN resource name
    can result in name collisions.  It's up to the caller
    to handle name collisions appropriately.

    """
    if not name:
        raise ValueError("Invalid name: %r" % name)
    word_separators = ['-', '_']
    for word_separator in word_separators:
        word_parts = [p for p in name.split(word_separator) if p]
        name = ''.join([w[0].upper() + w[1:] for w in word_parts])
    return re.sub(r'[^A-Za-z0-9]+', '', name)


def remove_stage_from_deployed_values(key, filename):
    # type: (str, str) -> None
    """Delete a top level key from the deployed JSON file."""
    final_values = {}  # type: Dict[str, Any]
    try:
        with open(filename, 'r') as f:
            final_values = json.load(f)
    except IOError:
        # If there is no file to delete from, then this funciton is a noop.
        return

    try:
        del final_values[key]
        with open(filename, 'wb') as outfile:
            data = serialize_to_json(final_values)
            outfile.write(data.encode('utf-8'))
    except KeyError:
        # If they key didn't exist then there is nothing to remove.
        pass


def record_deployed_values(deployed_values, filename):
    # type: (Dict[str, Any], str) -> None
    """Record deployed values to a JSON file.

    This allows subsequent deploys to lookup previously deployed values.

    """
    final_values = {}  # type: Dict[str, Any]
    if os.path.isfile(filename):
        with open(filename, 'r') as f:
            final_values = json.load(f)
    final_values.update(deployed_values)
    with open(filename, 'wb') as outfile:
        data = serialize_to_json(final_values)
        outfile.write(data.encode('utf-8'))


def serialize_to_json(data):
    # type: (Any) -> str
    """Serialize to pretty printed JSON.

    This includes using 2 space indentation, no trailing whitespace, and
    including a newline at the end of the JSON document.  Useful when you want
    to serialize JSON to disk.

    """
    return json.dumps(data, indent=2, separators=(',', ': ')) + '\n'


class ChaliceZipFile(zipfile.ZipFile):
    """Support deterministic zipfile generation.

    Normalizes datetime and permissions.

    """

    compression = 0  # Try to make mypy happy.
    _default_time_time = (1980, 1, 1, 0, 0, 0)

    def __init__(self, *args, **kwargs):
        # type: (Any, Any) -> None
        self._osutils = cast(OSUtils, kwargs.pop('osutils', OSUtils()))
        super(ChaliceZipFile, self).__init__(*args, **kwargs)

    # pylint: disable=W0221
    def write(self, filename, arcname=None, compress_type=None,
              compresslevel=None):
        # type: (StrPath, Optional[StrPath], OptInt, OptInt) -> None
        # Only supports files, py2.7 and 3 have different signatures.
        # We know that in our packager code we never call write() on
        # directories.
        zinfo = self._create_zipinfo(filename, arcname, compress_type)
        with open(filename, 'rb') as f:
            self.writestr(zinfo, f.read())

    def _create_zipinfo(self, filename, arcname, compress_type):
        # type: (StrPath, Optional[StrPath], Optional[int]) -> zipfile.ZipInfo
        # The main thing that prevents deterministic zip file generation
        # is that the mtime of the file is included in the zip metadata.
        # We don't actually care what the mtime is when we run on lambda,
        # so we always set it to the default value (which comes from
        # zipfile.py).  This ensures that as long as the file contents don't
        # change (or the permissions) then we'll always generate the exact
        # same zip file bytes.
        # We also can't use ZipInfo.from_file(), it's only in python3.
        st = self._osutils.stat(str(filename))
        if arcname is None:
            arcname = filename
        arcname = self._osutils.normalized_filename(str(arcname))
        arcname = arcname.lstrip(os.sep)
        zinfo = zipfile.ZipInfo(arcname, self._default_time_time)
        # The external_attr needs the upper 16 bits to be the file mode
        # so we have to shift it up to the right place.
        zinfo.external_attr = (st.st_mode & 0xFFFF) << 16
        zinfo.file_size = st.st_size
        zinfo.compress_type = compress_type or self.compression
        return zinfo


def create_zip_file(source_dir, outfile):
    # type: (str, str) -> None
    """Create a zip file from a source input directory.

    This function is intended to be an equivalent to
    `zip -r`.  You give it a source directory, `source_dir`,
    and it will recursively zip up the files into a zipfile
    specified by the `outfile` argument.

    """
    with ChaliceZipFile(outfile, 'w',
                        compression=zipfile.ZIP_DEFLATED,
                        osutils=OSUtils()) as z:
        for root, _, filenames in os.walk(source_dir):
            for filename in filenames:
                full_name = os.path.join(root, filename)
                archive_name = os.path.relpath(full_name, source_dir)
                z.write(full_name, archive_name)


class OSUtils(object):
    ZIP_DEFLATED = zipfile.ZIP_DEFLATED

    def environ(self):
        # type: () -> MutableMapping
        return os.environ

    def open(self, filename, mode):
        # type: (str, str) -> IO
        return open(filename, mode)

    def open_zip(self, filename, mode, compression=ZIP_DEFLATED):
        # type: (str, str, int) -> zipfile.ZipFile
        return ChaliceZipFile(filename, mode, compression=compression,
                              osutils=self)

    def remove_file(self, filename):
        # type: (str) -> None
        """Remove a file, noop if file does not exist."""
        # Unlike os.remove, if the file does not exist,
        # then this method does nothing.
        try:
            os.remove(filename)
        except OSError:
            pass

    def file_exists(self, filename):
        # type: (str) -> bool
        return os.path.isfile(filename)

    def get_file_contents(self, filename, binary=True, encoding='utf-8'):
        # type: (str, bool, Any) -> str
        # It looks like the type definition for io.open is wrong.
        # the encoding arg is unicode, but the actual type is
        # Optional[Text].  For now we have to use Any to keep mypy happy.
        if binary:
            mode = 'rb'
            # In binary mode the encoding is not used and most be None.
            encoding = None
        else:
            mode = 'r'
        with io.open(filename, mode, encoding=encoding) as f:
            return f.read()

    def set_file_contents(self, filename, contents, binary=True):
        # type: (str, str, bool) -> None
        if binary:
            mode = 'wb'
        else:
            mode = 'w'
        with open(filename, mode) as f:
            f.write(contents)

    def extract_zipfile(self, zipfile_path, unpack_dir):
        # type: (str, str) -> None
        with zipfile.ZipFile(zipfile_path, 'r') as z:
            z.extractall(unpack_dir)

    def extract_tarfile(self, tarfile_path, unpack_dir):
        # type: (str, str) -> None
        with tarfile.open(tarfile_path, 'r:*') as tar:
            tar.extractall(unpack_dir)

    def directory_exists(self, path):
        # type: (str) -> bool
        return os.path.isdir(path)

    def get_directory_contents(self, path):
        # type: (str) -> List[str]
        return os.listdir(path)

    def makedirs(self, path):
        # type: (str) -> None
        os.makedirs(path)

    def dirname(self, path):
        # type: (str) -> str
        return os.path.dirname(path)

    def abspath(self, path):
        # type: (str) -> str
        return os.path.abspath(path)

    def joinpath(self, *args):
        # type: (str) -> str
        return os.path.join(*args)

    def walk(self, path, followlinks=False):
        # type: (str, bool) -> Iterator[Tuple[str, List[str], List[str]]]
        return os.walk(path, followlinks=followlinks)

    def copytree(self, source, destination):
        # type: (str, str) -> None
        if not os.path.exists(destination):
            self.makedirs(destination)
        names = self.get_directory_contents(source)
        for name in names:
            new_source = os.path.join(source, name)
            new_destination = os.path.join(destination, name)
            if os.path.isdir(new_source):
                self.copytree(new_source, new_destination)
            else:
                shutil.copy2(new_source, new_destination)

    def rmtree(self, directory):
        # type: (str) -> None
        shutil.rmtree(directory)

    def copy(self, source, destination):
        # type: (str, str) -> None
        shutil.copy(source, destination)

    def move(self, source, destination):
        # type: (str, str) -> None
        shutil.move(source, destination)

    @contextlib.contextmanager
    def tempdir(self):
        # type: () -> Any
        tempdir = tempfile.mkdtemp()
        try:
            yield tempdir
        finally:
            shutil.rmtree(tempdir)

    def popen(self, command, stdout=None, stderr=None, env=None):
        # type: (List[str], OptInt, OptInt, EnvVars) -> subprocess.Popen
        p = subprocess.Popen(command, stdout=stdout, stderr=stderr, env=env)
        return p

    def mtime(self, path):
        # type: (str) -> float
        return os.stat(path).st_mtime

    def stat(self, path):
        # type: (str)  -> os.stat_result
        return os.stat(path)

    def normalized_filename(self, path):
        # type: (str) -> str
        """Normalize a path into a filename.

        This will normalize a file and remove any 'drive' component
        from the path on OSes that support drive specifications.

        """
        return os.path.normpath(os.path.splitdrive(path)[1])

    @property
    def pipe(self):
        # type: () -> int
        return subprocess.PIPE

    def basename(self, path):
        # type: (str) -> str
        return os.path.basename(path)


def getting_started_prompt(prompter):
    # type: (Any) -> bool
    return prompter.prompt(WELCOME_PROMPT)


class UI(object):
    def __init__(self, out=None, err=None, confirm=None):
        # type: (Optional[IO], Optional[IO], Any) -> None
        # I tried using a more exact type for the 'confirm'
        # param, but mypy seems to miss the 'if confirm is None'
        # check and types _confirm as Union[..., None].
        # So for now, we're using Any for this type.
        if out is None:
            out = sys.stdout
        if err is None:
            err = sys.stderr
        if confirm is None:
            confirm = click.confirm
        self._out = out
        self._err = err
        self._confirm = confirm

    def write(self, msg):
        # type: (str) -> None
        self._out.write(msg)

    def error(self, msg):
        # type: (str) -> None
        self._err.write(msg)

    def confirm(self, msg, default=False, abort=False):
        # type: (str, bool, bool) -> Any
        try:
            return self._confirm(msg, default, abort)
        except click.Abort:
            raise AbortedError()


class PipeReader(object):
    def __init__(self, stream):
        # type: (IO[bytes]) -> None
        self._stream = stream

    def read(self):
        # type: () -> OptBytes
        if not self._stream.isatty():
            return self._stream.read()
        return None


class TimestampConverter(object):
    # This is roughly based off of what's used in the AWS CLI.
    _RELATIVE_TIMESTAMP_REGEX = re.compile(
        r"(?P<amount>\d+)(?P<unit>s|m|h|d|w)$"
    )
    _TO_SECONDS = {
        's': 1,
        'm': 60,
        'h': 3600,
        'd': 24 * 3600,
        'w': 7 * 24 * 3600,
    }

    def __init__(self, now=None):
        # type: (Callable[[], datetime]) -> None
        if now is None:
            now = datetime.utcnow
        self._now = now

    def timestamp_to_datetime(self, timestamp):
        # type: (str) -> datetime
        """Convert a timestamp to a datetime object.

        This method detects what type of timestamp is provided and
        parse is appropriately to a timestamp object.

        """
        re_match = self._RELATIVE_TIMESTAMP_REGEX.match(timestamp)
        if re_match:
            datetime_value = self._relative_timestamp_to_datetime(
                int(re_match.group('amount')), re_match.group('unit')
            )
        else:
            datetime_value = self.parse_iso8601_timestamp(timestamp)
        return datetime_value

    def _relative_timestamp_to_datetime(self, amount, unit):
        # type: (int, str) -> datetime
        multiplier = self._TO_SECONDS[unit]
        return self._now() + timedelta(seconds=amount * multiplier * -1)

    def parse_iso8601_timestamp(self, timestamp):
        # type: (str) -> datetime
        return dateutil.parser.parse(timestamp, tzinfos={'GMT': tzutc()})



#!/usr/bin/env python
import os
from setuptools import setup, find_packages


with open('README.rst') as readme_file:
    README = readme_file.read()


def recursive_include(relative_dir):
    all_paths = []
    root_prefix = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), 'chalice')
    full_path = os.path.join(root_prefix, relative_dir)
    for rootdir, _, filenames in os.walk(full_path):
        for filename in filenames:
            abs_filename = os.path.join(rootdir, filename)
            all_paths.append(abs_filename[len(root_prefix) + 1:])
    return all_paths


install_requires = [
    'click>=7,<9.0',
    'botocore>=1.14.0,<2.0.0',
    'typing==3.6.4;python_version<"3.7"',
    'mypy-extensions==0.4.3',
    'six>=1.10.0,<2.0.0',
    'pip>=9,<21.4',
    'attrs>=19.3.0,<21.5.0',
    'jmespath>=0.9.3,<1.0.0',
    'pyyaml>=5.3.1,<7.0.0',
    'inquirer>=2.7.0,<3.0.0',
    'wheel',
    'setuptools'
]

setup(
    name='chalice',
    version='1.26.5',
    description="Microframework",
    long_description=README,
    author="James Saryerwinnie",
    author_email='js@jamesls.com',
    url='https://github.com/aws/chalice',
    packages=find_packages(exclude=['tests', 'tests.*']),
    install_requires=install_requires,
    extras_require={
        'event-file-poller': ['watchdog==0.9.0'],
        'cdk': [
            'aws_cdk.aws_iam>=1.85.0,<2.0',
            'aws_cdk.aws-s3-assets>=1.85.0,<2.0',
            'aws_cdk.cloudformation-include>=1.85.0,<2.0',
            'aws_cdk.core>=1.85.0,<2.0',
        ]
    },
    license="Apache License 2.0",
    package_data={'chalice': [
        '*.json', '*.pyi', 'py.typed'] + recursive_include('templates')},
    include_package_data=True,
    zip_safe=False,
    keywords='chalice',
    entry_points={
        'console_scripts': [
            'chalice = chalice.cli:main',
        ]
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        "Programming Language :: Python :: 3",
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
)



"""Control plane APIs for programatically building/deploying Chalice apps.

The eventual goal is to expose this as a public API that other tools can use
in their own integrations with Chalice, but this will need time for the APIs
to mature so for the time being this is an internal-only API.
"""
import os
from typing import Dict, Any
from chalice.cli.factory import CLIFactory


def package_app(project_dir, output_dir, stage, chalice_config=None,
                package_format='cloudformation', template_format='json'):
    # type: (str, str, str, Dict[str, Any], str, str) -> None
    factory = CLIFactory(project_dir, environ=os.environ)
    if chalice_config is None:
        chalice_config = {}
    config = factory.create_config_obj(
        stage, user_provided_params=chalice_config)
    options = factory.create_package_options()
    packager = factory.create_app_packager(config, options,
                                           package_format=package_format,
                                           template_format=template_format)
    packager.package_app(config, output_dir, stage)



import os
import sys
import json

from typing import Dict, Any, Optional, List, Union  # noqa
from chalice import __version__ as current_chalice_version
from chalice.app import Chalice  # noqa
from chalice.constants import DEFAULT_STAGE_NAME
from chalice.constants import DEFAULT_HANDLER_NAME


StrMap = Dict[str, Any]


class Config(object):
    """Configuration information for a chalice app.

    Configuration values for a chalice app can come from
    a number of locations, files on disk, CLI params, default
    values, etc.  This object is an abstraction that normalizes
    these values.

    In general, there's a precedence for looking up
    config values:

        * User specified params
        * Config file values
        * Default values

    A user specified parameter would mean values explicitly
    specified by a user.  Generally these come from command
    line parameters (e.g ``--profile prod``), but for the purposes
    of this object would also mean values passed explicitly to
    this config object when instantiated.

    Additionally, there are some configurations that can vary
    per chalice stage (note that a chalice stage is different
    from an api gateway stage).  For config values loaded from
    disk, we allow values to be specified for all stages or
    for a specific stage.  For example, take ``environment_variables``.
    You can set this as a top level key to specify env vars
    to set for all stages, or you can set this value per chalice
    stage to set stage-specific environment variables.  Consider
    this config file::

        {
          "environment_variables": {
            "TABLE": "foo"
          },
          "stages": {
            "dev": {
              "environment_variables": {
                "S3BUCKET": "devbucket"
              }
            },
            "prod": {
              "environment_variables": {
                "S3BUCKET": "prodbucket",
                "TABLE": "prodtable"
              }
            }
          }
        }

    If the currently configured chalice stage is "dev", then
    the config.environment_variables would be::

        {"TABLE": "foo", "S3BUCKET": "devbucket"}

    The "prod" stage would be::

        {"TABLE": "prodtable", "S3BUCKET": "prodbucket"}

    """

    def __init__(self,
                 chalice_stage=DEFAULT_STAGE_NAME,      # type: str
                 function_name=DEFAULT_HANDLER_NAME,    # type: str
                 user_provided_params=None,             # type: StrMap
                 config_from_disk=None,                 # type: StrMap
                 default_params=None,                   # type: StrMap
                 layers=None,                           # type: List[str]
                 ):
        # type: (...) -> None
        #: Params that a user provided explicitly,
        #: typically via the command line.
        self.chalice_stage = chalice_stage
        self.function_name = function_name
        if user_provided_params is None:
            user_provided_params = {}
        self._user_provided_params = user_provided_params
        #: The json.loads() from .chalice/config.json
        if config_from_disk is None:
            config_from_disk = {}
        self._config_from_disk = config_from_disk
        if default_params is None:
            default_params = {}
        self._default_params = default_params
        self._chalice_app = None
        self._layers = layers

    @classmethod
    def create(cls, chalice_stage=DEFAULT_STAGE_NAME,
               function_name=DEFAULT_HANDLER_NAME,
               **kwargs):
        # type: (str, str, **Any) -> Config
        return cls(chalice_stage=chalice_stage,
                   user_provided_params=kwargs.copy())

    @property
    def profile(self):
        # type: () -> str
        return self._chain_lookup('profile')

    @property
    def app_name(self):
        # type: () -> str
        return self._chain_lookup('app_name')

    @property
    def project_dir(self):
        # type: () -> str
        return self._chain_lookup('project_dir')

    @property
    def chalice_app(self):
        # type: () -> Chalice
        v = self._chain_lookup('chalice_app')
        # There's two value we support.  If the value
        # is a chalice app, we return it as is.
        # Otherwise, we assume it's a callable that creates
        # a chalice app.  This is used to lazy load the chalice
        # app.
        if isinstance(v, Chalice):
            return v
        elif self._chalice_app is not None:
            return self._chalice_app
        elif not callable(v):
            raise TypeError("Unable to load chalice app, lazy loader is "
                            "not callable: %s" % v)
        app = v()
        self._chalice_app = app
        # Keep mypy happy.
        return app

    @property
    def config_from_disk(self):
        # type: () -> StrMap
        return self._config_from_disk

    @property
    def lambda_python_version(self):
        # type: () -> str
        # We may open this up to configuration later, but for now,
        # we attempt to match your python version to the closest version
        # supported by lambda.
        major, minor = sys.version_info[0], sys.version_info[1]
        if major == 2:
            return 'python2.7'
        # Python 3 for backwards compatibility needs to select python3.6
        # for python versions 3.0-3.6. 3.7 and higher will use python3.7.
        elif (major, minor) <= (3, 6):
            return 'python3.6'
        elif (major, minor) <= (3, 7):
            return 'python3.7'
        elif (major, minor) <= (3, 8):
            return 'python3.8'
        return 'python3.9'

    @property
    def layers(self):
        # type: () -> List
        return self._chain_lookup('layers',
                                  varies_per_chalice_stage=True,
                                  varies_per_function=True)

    @property
    def api_gateway_custom_domain(self):
        # type: () -> StrMap
        return self._chain_lookup('api_gateway_custom_domain',
                                  varies_per_chalice_stage=True)

    @property
    def websocket_api_custom_domain(self):
        # type: () -> StrMap
        return self._chain_lookup('websocket_api_custom_domain',
                                  varies_per_chalice_stage=True)

    def _chain_lookup(self, name, varies_per_chalice_stage=False,
                      varies_per_function=False):
        # type: (str, bool, bool) -> Any
        search_dicts = [self._user_provided_params]
        if varies_per_chalice_stage:
            search_dicts.append(
                self._config_from_disk.get('stages', {}).get(
                    self.chalice_stage, {}))
        if varies_per_function:
            # search order:
            # config['stages']['lambda_functions']
            # config['stages']
            # config['lambda_functions']
            search_dicts.insert(
                0, self._config_from_disk.get('stages', {}).get(
                    self.chalice_stage, {}).get('lambda_functions', {}).get(
                        self.function_name, {}))
            search_dicts.append(
                self._config_from_disk.get('lambda_functions', {}).get(
                    self.function_name, {}))
        search_dicts.extend([self._config_from_disk, self._default_params])
        for cfg_dict in search_dicts:
            if isinstance(cfg_dict, dict) and cfg_dict.get(name) is not None:
                return cfg_dict[name]

    def _chain_merge(self, name):
        # type: (str) -> Dict[str, Any]
        # Merge values for all search dicts instead of returning on first
        # found.
        search_dicts = [
            # This is reverse order to _chain_lookup().
            self._default_params,
            self._config_from_disk,
            self._config_from_disk.get('stages', {}).get(
                self.chalice_stage, {}),
            self._config_from_disk.get('stages', {}).get(
                self.chalice_stage, {}).get('lambda_functions', {}).get(
                    self.function_name, {}),
            self._user_provided_params,
        ]
        final = {}
        for cfg_dict in search_dicts:
            value = cfg_dict.get(name, {})
            if isinstance(value, dict):
                final.update(value)
        return final

    @property
    def config_file_version(self):
        # type: () -> str
        return self._config_from_disk.get('version', '1.0')

    # These are all config values that can vary per
    # chalice stage.

    @property
    def api_gateway_stage(self):
        # type: () -> str
        return self._chain_lookup('api_gateway_stage',
                                  varies_per_chalice_stage=True)

    @property
    def api_gateway_endpoint_type(self):
        # type: () -> str
        return self._chain_lookup('api_gateway_endpoint_type',
                                  varies_per_chalice_stage=True)

    @property
    def api_gateway_endpoint_vpce(self):
        # type: () -> Union[str, List[str]]
        return self._chain_lookup('api_gateway_endpoint_vpce',
                                  varies_per_chalice_stage=True)

    @property
    def api_gateway_policy_file(self):
        # type: () -> str
        return self._chain_lookup('api_gateway_policy_file',
                                  varies_per_chalice_stage=True)

    @property
    def minimum_compression_size(self):
        # type: () -> int
        return self._chain_lookup('minimum_compression_size',
                                  varies_per_chalice_stage=True)

    @property
    def iam_policy_file(self):
        # type: () -> str
        return self._chain_lookup('iam_policy_file',
                                  varies_per_chalice_stage=True,
                                  varies_per_function=True)

    @property
    def lambda_memory_size(self):
        # type: () -> int
        return self._chain_lookup('lambda_memory_size',
                                  varies_per_chalice_stage=True,
                                  varies_per_function=True)

    @property
    def lambda_timeout(self):
        # type: () -> int
        return self._chain_lookup('lambda_timeout',
                                  varies_per_chalice_stage=True,
                                  varies_per_function=True)

    @property
    def automatic_layer(self):
        # type: () -> bool
        v = self._chain_lookup('automatic_layer',
                               varies_per_chalice_stage=True,
                               varies_per_function=False)
        if v is None:
            return False
        return v

    @property
    def iam_role_arn(self):
        # type: () -> str
        return self._chain_lookup('iam_role_arn',
                                  varies_per_chalice_stage=True,
                                  varies_per_function=True)

    @property
    def manage_iam_role(self):
        # type: () -> bool
        result = self._chain_lookup('manage_iam_role',
                                    varies_per_chalice_stage=True,
                                    varies_per_function=True)
        if result is None:
            # To simplify downstream code, if manage_iam_role
            # is None (indicating the user hasn't configured/specified this
            # value anywhere), then we'll return a default value of True.
            # Otherwise client code has to do an awkward
            # "if manage_iam_role is None and not manage_iam_role".
            return True
        return result

    @property
    def autogen_policy(self):
        # type: () -> bool
        return self._chain_lookup('autogen_policy',
                                  varies_per_chalice_stage=True,
                                  varies_per_function=True)

    @property
    def xray_enabled(self):
        # type: () -> bool
        return self._chain_lookup('xray',
                                  varies_per_chalice_stage=True,
                                  varies_per_function=True)

    @property
    def environment_variables(self):
        # type: () -> Dict[str, str]
        return self._chain_merge('environment_variables')

    @property
    def tags(self):
        # type: () -> Dict[str, str]
        tags = self._chain_merge('tags')
        tags['aws-chalice'] = 'version=%s:stage=%s:app=%s' % (
            current_chalice_version, self.chalice_stage, self.app_name)
        return tags

    @property
    def security_group_ids(self):
        # type: () -> List[str]
        return self._chain_lookup('security_group_ids',
                                  varies_per_chalice_stage=True,
                                  varies_per_function=True)

    @property
    def subnet_ids(self):
        # type: () -> List[str]
        return self._chain_lookup('subnet_ids',
                                  varies_per_chalice_stage=True,
                                  varies_per_function=True)

    @property
    def reserved_concurrency(self):
        # type: () -> int
        return self._chain_lookup('reserved_concurrency',
                                  varies_per_chalice_stage=True,
                                  varies_per_function=True)

    def scope(self, chalice_stage, function_name):
        # type: (str, str) -> Config
        # Used to create a new config object that's scoped to a different
        # stage and/or function.  This creates a completely separate copy.
        # This is preferred over mutating the existing config obj.
        # We technically don't need to do a copy here, but this avoids
        # any possible issues if we ever mutate the config values.
        clone = self.__class__(
            chalice_stage=chalice_stage,
            function_name=function_name,
            user_provided_params=self._user_provided_params,
            config_from_disk=self._config_from_disk,
            default_params=self._default_params,
        )
        return clone

    def deployed_resources(self, chalice_stage_name):
        # type: (str) -> DeployedResources
        """Return resources associated with a given stage.

        If a deployment to a given stage has never happened,
        this method will return a value of None.

        """
        # This is arguably the wrong level of abstraction.
        # We might be able to move this elsewhere.
        deployed_file = os.path.join(
            self.project_dir, '.chalice', 'deployed',
            '%s.json' % chalice_stage_name)
        data = self._load_json_file(deployed_file)
        if data is not None:
            schema_version = data.get('schema_version', '1.0')
            if schema_version != '2.0':
                raise ValueError("Unsupported schema version (%s) in file: %s"
                                 % (schema_version, deployed_file))
            return DeployedResources(data)
        return self._try_old_deployer_values(chalice_stage_name)

    def _try_old_deployer_values(self, chalice_stage_name):
        # type: (str) -> DeployedResources
        # They are upgrading from v1.0 to v2.0 of the deployed.json
        # schema.  Attempt to auto convert for them.
        old_deployed_file = os.path.join(self.project_dir, '.chalice',
                                         'deployed.json')
        data = self._load_json_file(old_deployed_file)
        if data is None or chalice_stage_name not in data:
            return DeployedResources.empty()
        return self._upgrade_deployed_values(chalice_stage_name, data)

    def _load_json_file(self, deployed_file):
        # type: (str) -> Any
        if not os.path.isfile(deployed_file):
            return None
        with open(deployed_file, 'r') as f:
            return json.load(f)

    def _upgrade_deployed_values(self, chalice_stage_name, data):
        # type: (str, Any) -> DeployedResources
        deployed = data[chalice_stage_name]
        prefix = '%s-%s-' % (self.app_name, chalice_stage_name)
        resources = []  # type: List[Dict[str, Any]]
        self._upgrade_lambda_functions(resources, deployed, prefix)
        self._upgrade_rest_api(resources, deployed)
        return DeployedResources(
            {'resources': resources, 'schema_version': '2.0'})

    def _upgrade_lambda_functions(self, resources, deployed, prefix):
        # type: (List[Dict[str, Any]], Dict[str, Any], str) -> None
        lambda_functions = deployed.get('lambda_functions', {})
        # In chalice 0.10.0, the lambda_functions had the format
        # {"function-name": "lambda_arn"} as opposed to
        # {"function-name": {"arn": "lambda_arn", "type": "...'}} used
        # in later versions of chalice.  We'll check for both cases
        # so people can upgrade from 0.10.0 to the new deployer.
        is_pre_10_format = not all(
            isinstance(v, dict)
            for v in lambda_functions.values()
        )
        if is_pre_10_format:
            lambda_functions = {
                # The only supported lambda functions in 0.10.0
                # was built in authorizers.
                k: {'type': 'authorizer', 'arn': v}
                for k, v in lambda_functions.items()
            }
        for name, values in lambda_functions.items():
            short_name = name[len(prefix):]
            current = {
                'resource_type': 'lambda_function',
                'lambda_arn': values['arn'],
                'name': short_name,
            }
            resources.append(current)

    def _upgrade_rest_api(self, resources, deployed):
        # type: (List[Dict[str, Any]], Dict[str, Any]) -> None
        resources.extend([
            {'name': 'api_handler',
             'resource_type': 'lambda_function',
             'lambda_arn': deployed['api_handler_arn']},
            {'name': 'rest_api',
             'resource_type': 'rest_api',
             'rest_api_id': deployed['rest_api_id']},
        ])


class DeployedResources(object):
    def __init__(self, deployed_values):
        # type: (Dict[str, Any]) -> None
        self._deployed_values = deployed_values['resources']
        self._deployed_values_by_name = {
            resource['name']: resource
            for resource in deployed_values['resources']
        }

    @classmethod
    def empty(cls):
        # type: () -> DeployedResources
        return cls({'resources': [], 'schema_version': '2.0'})

    def resource_values(self, name):
        # type: (str) -> Dict[str, Any]
        if 'api_mapping' in name:
            name = name.split('.')[0]

        try:
            return self._deployed_values_by_name[name]
        except KeyError:
            raise ValueError("Resource does not exist: %s" % name)

    def resource_names(self):
        # type: () -> List[str]
        return [r['name'] for r in self._deployed_values]



"""Chalice deployer module.

The deployment system in chalice is broken down into a pipeline of multiple
stages.  Each stage takes the input and transforms it to some other form.  The
reason for this is so that each stage can stay simple and focused on only a
single part of the deployment process.  This makes the code easier to follow
and easier to test.  The biggest downside is that adding support for a new
resource type is split across several objects now, but I imagine as we add
support for more resource types, we'll see common patterns emerge that we can
extract out into higher levels of abstraction.

These are the stages of the deployment process.

Application Graph Builder
=========================

The first stage is the resource graph builder.  This takes the objects in the
``Chalice`` app and structures them into an ``Application`` object which
consists of various ``models.Model`` objects.  These models are just python
objects that describe the attributes of various AWS resources.  These
models don't have any behavior on their own.

Dependency Builder
==================

This process takes the graph of resources created from the previous step and
orders them such that all objects are listed before objects that depend on
them.  The AWS resources in the ``chalice.deploy.models`` module also model
their required dependencies (see the ``dependencies()`` methods of the models).
This is the mechanism that's used to build the correct dependency ordering.

Local Build Stage
=================

This takes the ordered list of resources and allows any local build processes
to occur.  The rule of thumb here is no remote AWS calls.  This stage includes
auto policy generation, pip packaging, injecting default values, etc.  To
clarify which attributes are affected by the build stage, they'll usually have
a value of ``models.Placeholder.BUILD_STAGE``.  Processors in the build stage
will replaced those ``models.Placeholder.BUILD_STAGE`` values with whatever the
"built" value is (e.g the filename of the zipped deployment package).

For example, we know when we create a lambda function that we need to create a
deployment package, but we don't know the name nor contents of the deployment
package until the ``LambdaDeploymentPackager`` runs.  Therefore, the Resource
Builder stage can record the fact that it knows that a
``models.DeploymentPackage`` is needed, but use
``models.Placeholder.BUILD_STAGE`` for the value of the filename.  The enum
values aren't strictly necessary, they just add clarity about when this value
is expected to be filled in.  These could also just be set to ``None`` and be
of type ``Optional[T]``.


Execution Plan Stage
====================

This stage takes the ordered list of resources and figures out what AWS API
calls we have to make.  For example, if a resource doesn't exist at all, we'll
need to make a ``create_*`` call.  If the resource exists, we may need to make
a series of ``update_*`` calls.  If the resource exists and is already up to
date, we might not need to make any calls at all.  The output of this stage is
a list of ``APICall`` objects.  This stage doesn't actually make the mutating
API calls, it only figures out what calls we should make.  This stage will
typically only make ``describe/list`` AWS calls.


The Executor
============

This takes the list of ``APICall`` objects from the previous stage and finally
executes them.  It also manages taking the output of API calls and storing them
in variables so they can be referenced in subsequent ``APICall`` objects (see
the ``Variable`` class to see how this is used).  For example, if a lambda
function needs the ``role_arn`` that's the result of a previous ``create_role``
API call, a ``Variable`` object is used to forward this information.

The executor also records these variables with their associated resources so a
``deployed.json`` file can be written to disk afterwards.  An ``APICall``
takes an optional resource object when it's created whose ``resource_name``
is used as the key in the ``deployed.json`` dictionary.


"""
# pylint: disable=too-many-lines
import json
import textwrap
import socket
import logging

import botocore.exceptions
from botocore.vendored.requests import ConnectionError as \
    RequestsConnectionError
from botocore.session import Session  # noqa
from typing import Optional, Dict, List, Any, Type, cast  # noqa

from chalice.config import Config  # noqa
from chalice.compat import is_broken_pipe_error
from chalice.awsclient import DeploymentPackageTooLargeError
from chalice.awsclient import LambdaClientError
from chalice.awsclient import AWSClientError
from chalice.awsclient import TypedAWSClient
from chalice.constants import MAX_LAMBDA_DEPLOYMENT_SIZE
from chalice.constants import VPC_ATTACH_POLICY
from chalice.constants import DEFAULT_LAMBDA_TIMEOUT
from chalice.constants import DEFAULT_LAMBDA_MEMORY_SIZE
from chalice.constants import DEFAULT_TLS_VERSION
from chalice.constants import SQS_EVENT_SOURCE_POLICY
from chalice.constants import KINESIS_EVENT_SOURCE_POLICY
from chalice.constants import DDB_EVENT_SOURCE_POLICY
from chalice.constants import POST_TO_WEBSOCKET_CONNECTION_POLICY
from chalice.deploy import models
from chalice.deploy.appgraph import ApplicationGraphBuilder, DependencyBuilder
from chalice.deploy.executor import BaseExecutor  # noqa
from chalice.deploy.executor import Executor
from chalice.deploy.executor import DisplayOnlyExecutor
from chalice.deploy.packager import PipRunner
from chalice.deploy.packager import SubprocessPip
from chalice.deploy.packager import DependencyBuilder as PipDependencyBuilder
from chalice.deploy.packager import LambdaDeploymentPackager
from chalice.deploy.packager import AppOnlyDeploymentPackager
from chalice.deploy.packager import LayerDeploymentPackager
from chalice.deploy.packager import BaseLambdaDeploymentPackager  # noqa
from chalice.deploy.packager import EmptyPackageError
from chalice.deploy.planner import PlanStage
from chalice.deploy.planner import RemoteState
from chalice.deploy.planner import NoopPlanner
from chalice.deploy.swagger import TemplatedSwaggerGenerator
from chalice.deploy.swagger import SwaggerGenerator  # noqa
from chalice.deploy.sweeper import ResourceSweeper
from chalice.deploy.validate import validate_configuration
from chalice.policy import AppPolicyGenerator
from chalice.utils import OSUtils
from chalice.utils import UI
from chalice.utils import serialize_to_json


OptStr = Optional[str]
LOGGER = logging.getLogger(__name__)


_AWSCLIENT_EXCEPTIONS = (
    botocore.exceptions.ClientError, AWSClientError
)


class ChaliceDeploymentError(Exception):
    def __init__(self, error):
        # type: (Exception) -> None
        self.original_error = error
        where = self._get_error_location(error)
        msg = self._wrap_text(
            'ERROR - %s, received the following error:' % where
        )
        msg += '\n\n'
        msg += self._wrap_text(self._get_error_message(error), indent=' ')
        msg += '\n\n'
        suggestion = self._get_error_suggestion(error)
        if suggestion is not None:
            msg += self._wrap_text(suggestion)
        super(ChaliceDeploymentError, self).__init__(msg)

    def _get_error_location(self, error):
        # type: (Exception) -> str
        where = 'While deploying your chalice application'
        if isinstance(error, LambdaClientError):
            where = (
                'While sending your chalice handler code to Lambda to %s '
                'function "%s"' % (
                    self._get_verb_from_client_method(
                        error.context.client_method_name),
                    error.context.function_name
                )
            )
        return where

    def _get_error_message(self, error):
        # type: (Exception) -> str
        msg = str(error)
        if isinstance(error, LambdaClientError):
            if isinstance(error.original_error, RequestsConnectionError):
                msg = self._get_error_message_for_connection_error(
                    error.original_error)
        return msg

    def _get_error_message_for_connection_error(self, connection_error):
        # type: (RequestsConnectionError) -> str

        # To get the underlying error that raised the
        # requests.ConnectionError it is required to go down two levels of
        # arguments to get the underlying exception. The instantiation of
        # one of these exceptions looks like this:
        #
        # requests.ConnectionError(
        #     urllib3.exceptions.ProtocolError(
        #         'Connection aborted.', <SomeException>)
        # )
        message = connection_error.args[0].args[0]
        underlying_error = connection_error.args[0].args[1]
        if is_broken_pipe_error(underlying_error):
            message += (
                ' Lambda closed the connection before chalice finished '
                'sending all of the data.'
            )
        elif isinstance(underlying_error, socket.timeout):
            message += ' Timed out sending your app to Lambda.'
        return message

    def _get_error_suggestion(self, error):
        # type: (Exception) -> OptStr
        suggestion = None
        if isinstance(error, DeploymentPackageTooLargeError):
            suggestion = (
                'To avoid this error, decrease the size of your chalice '
                'application by removing code or removing '
                'dependencies from your chalice application.'
            )
            deployment_size = error.context.deployment_size
            if deployment_size > MAX_LAMBDA_DEPLOYMENT_SIZE:
                size_warning = (
                    'This is likely because the deployment package is %s. '
                    'Lambda only allows deployment packages that are %s or '
                    'less in size.' % (
                        self._get_mb(deployment_size),
                        self._get_mb(MAX_LAMBDA_DEPLOYMENT_SIZE)
                    )
                )
                suggestion = size_warning + ' ' + suggestion
        return suggestion

    def _wrap_text(self, text, indent=''):
        # type: (str, str) -> str
        return '\n'.join(
            textwrap.wrap(
                text, 79, replace_whitespace=False, drop_whitespace=False,
                initial_indent=indent, subsequent_indent=indent
            )
        )

    def _get_verb_from_client_method(self, client_method_name):
        # type: (str) -> str
        client_method_name_to_verb = {
            'update_function_code': 'update',
            'create_function': 'create'
        }
        return client_method_name_to_verb.get(
            client_method_name, client_method_name)

    def _get_mb(self, value):
        # type: (int) -> str
        return '%.1f MB' % (float(value) / (1024 ** 2))


def create_plan_only_deployer(session, config, ui):
    # type: (Session, Config, UI) -> Deployer
    return _create_deployer(session, config, ui, DisplayOnlyExecutor,
                            NoopResultsRecorder)


def create_default_deployer(session, config, ui):
    # type: (Session, Config, UI) -> Deployer
    return _create_deployer(session, config, ui, Executor, ResultsRecorder)


def _create_deployer(session,       # type: Session
                     config,        # type: Config
                     ui,            # type: UI
                     executor_cls,  # type: Type[BaseExecutor]
                     recorder_cls,  # type: Type[ResultsRecorder]
                     ):
    # type: (...) -> Deployer
    client = TypedAWSClient(session)
    osutils = OSUtils()
    return Deployer(
        application_builder=ApplicationGraphBuilder(),
        deps_builder=DependencyBuilder(),
        build_stage=create_build_stage(
            osutils, UI(), TemplatedSwaggerGenerator(), config
        ),
        plan_stage=PlanStage(
            osutils=osutils, remote_state=RemoteState(
                client, config.deployed_resources(config.chalice_stage)),
        ),
        sweeper=ResourceSweeper(),
        executor=executor_cls(client, ui),
        recorder=recorder_cls(osutils=osutils),
    )


def create_build_stage(osutils, ui, swagger_gen, config):
    # type: (OSUtils, UI, SwaggerGenerator, Config) -> BuildStage
    pip_runner = PipRunner(pip=SubprocessPip(osutils=osutils),
                           osutils=osutils)
    dependency_builder = PipDependencyBuilder(
        osutils=osutils,
        pip_runner=pip_runner
    )
    deployment_packager = cast(BaseDeployStep, None)
    if config.automatic_layer:
        deployment_packager = ManagedLayerDeploymentPackager(
            lambda_packager=AppOnlyDeploymentPackager(
                osutils=osutils,
                dependency_builder=dependency_builder,
                ui=ui,
            ),
            layer_packager=LayerDeploymentPackager(
                osutils=osutils,
                dependency_builder=dependency_builder,
                ui=ui,
            )
        )
    else:
        deployment_packager = DeploymentPackager(
            packager=LambdaDeploymentPackager(
                osutils=osutils,
                dependency_builder=dependency_builder,
                ui=ui,
            )
        )
    build_stage = BuildStage(
        steps=[
            InjectDefaults(),
            deployment_packager,
            PolicyGenerator(
                policy_gen=AppPolicyGenerator(
                    osutils=osutils
                ),
                osutils=osutils,
            ),
            SwaggerBuilder(
                swagger_generator=swagger_gen,
            ),
            LambdaEventSourcePolicyInjector(),
            WebsocketPolicyInjector()
        ],
    )
    return build_stage


def create_deletion_deployer(client, ui):
    # type: (TypedAWSClient, UI) -> Deployer
    return Deployer(
        application_builder=ApplicationGraphBuilder(),
        deps_builder=DependencyBuilder(),
        build_stage=BuildStage(steps=[]),
        plan_stage=NoopPlanner(),
        sweeper=ResourceSweeper(),
        executor=Executor(client, ui),
        recorder=ResultsRecorder(osutils=OSUtils()),
    )


class Deployer(object):
    BACKEND_NAME = 'api'

    def __init__(self,
                 application_builder,  # type: ApplicationGraphBuilder
                 deps_builder,         # type: DependencyBuilder
                 build_stage,          # type: BuildStage
                 plan_stage,           # type: PlanStage
                 sweeper,              # type: ResourceSweeper
                 executor,             # type: BaseExecutor
                 recorder,             # type: ResultsRecorder
                 ):
        # type: (...) -> None
        self._application_builder = application_builder
        self._deps_builder = deps_builder
        self._build_stage = build_stage
        self._plan_stage = plan_stage
        self._sweeper = sweeper
        self._executor = executor
        self._recorder = recorder

    def deploy(self, config, chalice_stage_name):
        # type: (Config, str) -> Dict[str, Any]
        try:
            return self._deploy(config, chalice_stage_name)
        except _AWSCLIENT_EXCEPTIONS as e:
            raise ChaliceDeploymentError(e)

    def _deploy(self, config, chalice_stage_name):
        # type: (Config, str) -> Dict[str, Any]
        self._validate_config(config)
        application = self._application_builder.build(
            config, chalice_stage_name)
        resources = self._deps_builder.build_dependencies(application)
        self._build_stage.execute(config, resources)
        # Rebuild dependencies in case the build stage modified
        # the app graph.
        resources = self._deps_builder.build_dependencies(application)
        plan = self._plan_stage.execute(resources)
        self._sweeper.execute(plan, config)
        self._executor.execute(plan)
        deployed_values = {
            'resources': self._executor.resource_values,
            'schema_version': '2.0',
            'backend': self.BACKEND_NAME,
        }
        self._recorder.record_results(
            deployed_values,
            chalice_stage_name,
            config.project_dir,
        )
        return deployed_values

    def _validate_config(self, config):
        # type: (Config) -> None
        try:
            validate_configuration(config)
        except ValueError as e:
            raise ChaliceDeploymentError(e)


class BaseDeployStep(object):
    def handle(self, config, resource):
        # type: (Config, models.Model) -> None
        name = 'handle_%s' % resource.__class__.__name__.lower()
        handler = getattr(self, name, None)
        if handler is not None:
            handler(config, resource)


class InjectDefaults(BaseDeployStep):
    def __init__(self, lambda_timeout=DEFAULT_LAMBDA_TIMEOUT,
                 lambda_memory_size=DEFAULT_LAMBDA_MEMORY_SIZE,
                 tls_version=DEFAULT_TLS_VERSION):
        # type: (int, int, str) -> None
        self._lambda_timeout = lambda_timeout
        self._lambda_memory_size = lambda_memory_size
        self._tls_version = DEFAULT_TLS_VERSION

    def handle_lambdafunction(self, config, resource):
        # type: (Config, models.LambdaFunction) -> None
        if resource.timeout is None:
            resource.timeout = self._lambda_timeout
        if resource.memory_size is None:
            resource.memory_size = self._lambda_memory_size

    def handle_domainname(self, config, resource):
        # type: (Config, models.DomainName) -> None
        if resource.tls_version is None:
            resource.tls_version = models.TLSVersion.create(
                DEFAULT_TLS_VERSION)


class DeploymentPackager(BaseDeployStep):
    def __init__(self, packager):
        # type: (LambdaDeploymentPackager) -> None
        self._packager = packager

    def handle_deploymentpackage(self, config, resource):
        # type: (Config, models.DeploymentPackage) -> None
        if isinstance(resource.filename, models.Placeholder):
            zip_filename = self._packager.create_deployment_package(
                config.project_dir, config.lambda_python_version)
            resource.filename = zip_filename


class ManagedLayerDeploymentPackager(BaseDeployStep):
    # If we're creating a layer for non-app code there's two different
    # packagers we need.  One for the Lambda functions (app code) and
    # one for the Lambda layer (requirements.txt + vendor).
    def __init__(self,
                 lambda_packager,  # type: BaseLambdaDeploymentPackager
                 layer_packager,   # type: BaseLambdaDeploymentPackager
                 ):
        # type: (...) -> None
        self._lambda_packager = lambda_packager
        self._layer_packager = layer_packager

    def handle_lambdafunction(self, config, resource):
        # type: (Config, models.LambdaFunction) -> None
        if isinstance(resource.deployment_package.filename,
                      models.Placeholder):
            zip_filename = self._lambda_packager.create_deployment_package(
                config.project_dir, config.lambda_python_version
            )
            resource.deployment_package.filename = zip_filename
        if resource.managed_layer is not None and \
                resource.managed_layer.is_empty:
            # Lambda doesn't allow us to create an empty layer so if we've
            # tried to create the deployment package and determined it's empty
            # we should remove the managed layer from the model entirely so
            # downstream consumers don't have to worry about it.
            resource.managed_layer = None

    def handle_lambdalayer(self, config, resource):
        # type: (Config, models.LambdaLayer) -> None
        if isinstance(resource.deployment_package.filename,
                      models.Placeholder):
            try:
                zip_filename = self._layer_packager.create_deployment_package(
                    config.project_dir, config.lambda_python_version
                )
                resource.deployment_package.filename = zip_filename
            except EmptyPackageError:
                # An empty package is not valid in Lambda.  There's
                # no point in trying to represent it in the model
                # because nothing downstream (planner, SAM/tf) can
                # consume as a useful entity.
                resource.is_empty = True


class SwaggerBuilder(BaseDeployStep):
    def __init__(self, swagger_generator):
        # type: (SwaggerGenerator) -> None
        self._swagger_generator = swagger_generator

    def handle_restapi(self, config, resource):
        # type: (Config, models.RestAPI) -> None
        swagger_doc = self._swagger_generator.generate_swagger(
            config.chalice_app, resource)
        resource.swagger_doc = swagger_doc


class LambdaEventSourcePolicyInjector(BaseDeployStep):
    def __init__(self):
        # type: () -> None
        self._sqs_policy_injected = False
        self._kinesis_policy_injected = False
        self._ddb_policy_injected = False

    def handle_sqseventsource(self, config, resource):
        # type: (Config, models.SQSEventSource) -> None
        # The sqs integration works by polling for
        # available records so the lambda function needs
        # permission to call sqs.
        role = resource.lambda_function.role
        if not self._sqs_policy_injected and \
                self._needs_policy_injected(role):
            # mypy can't follow the type narrowing from
            # _needs_policy_injected so we're working around
            # that by explicitly casting the role.
            role = cast(models.ManagedIAMRole, role)
            document = cast(Dict[str, Any], role.policy.document)
            self._inject_trigger_policy(document,
                                        SQS_EVENT_SOURCE_POLICY.copy())
            self._sqs_policy_injected = True

    def handle_kinesiseventsource(self, config, resource):
        # type: (Config, models.KinesisEventSource) -> None
        role = resource.lambda_function.role
        if not self._kinesis_policy_injected and \
                self._needs_policy_injected(role):
            # See commen in handle_kinesiseventsource about this cast.
            role = cast(models.ManagedIAMRole, role)
            document = cast(Dict[str, Any], role.policy.document)
            self._inject_trigger_policy(document,
                                        KINESIS_EVENT_SOURCE_POLICY.copy())
            self._kinesis_policy_injected = True

    def handle_dynamodbeventsource(self, config, resource):
        # type: (Config, models.KinesisEventSource) -> None
        role = resource.lambda_function.role
        if not self._ddb_policy_injected and \
                self._needs_policy_injected(role):
            # See commen in handle_kinesiseventsource about this cast.
            role = cast(models.ManagedIAMRole, role)
            document = cast(Dict[str, Any], role.policy.document)
            self._inject_trigger_policy(document,
                                        DDB_EVENT_SOURCE_POLICY.copy())
            self._ddb_policy_injected = True

    def _needs_policy_injected(self, role):
        # type: (models.IAMRole) -> bool
        return (
            isinstance(role, models.ManagedIAMRole) and
            isinstance(role.policy, models.AutoGenIAMPolicy) and
            not isinstance(role.policy.document, models.Placeholder)
        )

    def _inject_trigger_policy(self, document, policy):
        # type: (Dict[str, Any], Dict[str, Any]) -> None
        document['Statement'].append(policy)


class WebsocketPolicyInjector(BaseDeployStep):
    def __init__(self):
        # type: () -> None
        self._policy_injected = False

    def handle_websocketapi(self, config, resource):
        # type: (Config, models.WebsocketAPI) -> None
        self._inject_into_function(config, resource.connect_function)
        self._inject_into_function(config, resource.message_function)
        self._inject_into_function(config, resource.disconnect_function)

    def _inject_into_function(self, config, lambda_function):
        # type: (Config, Optional[models.LambdaFunction]) -> None
        if lambda_function is None:
            return
        role = lambda_function.role
        if role is None:
            return
        if (not self._policy_injected and
            isinstance(role, models.ManagedIAMRole) and
            isinstance(role.policy, models.AutoGenIAMPolicy) and
            not isinstance(role.policy.document,
                           models.Placeholder)):
            self._inject_policy(
                role.policy.document,
                POST_TO_WEBSOCKET_CONNECTION_POLICY.copy())
        self._policy_injected = True

    def _inject_policy(self, document, policy):
        # type: (Dict[str, Any], Dict[str, Any]) -> None
        document['Statement'].append(policy)


class PolicyGenerator(BaseDeployStep):
    def __init__(self, policy_gen, osutils):
        # type: (AppPolicyGenerator, OSUtils) -> None
        self._policy_gen = policy_gen
        self._osutils = osutils

    def _read_document_from_file(self, filename):
        # type: (PolicyGenerator, str) -> Dict[str, Any]
        try:
            return json.loads(self._osutils.get_file_contents(filename))
        except IOError as e:
            raise RuntimeError("Unable to load IAM policy file %s: %s"
                               % (filename, e))

    def handle_filebasediampolicy(self, config, resource):
        # type: (Config, models.FileBasedIAMPolicy) -> None
        resource.document = self._read_document_from_file(resource.filename)

    def handle_restapi(self, config, resource):
        # type: (Config, models.RestAPI) -> None
        if resource.policy and isinstance(
                resource.policy, models.FileBasedIAMPolicy):
            resource.policy.document = self._read_document_from_file(
                resource.policy.filename)

    def handle_autogeniampolicy(self, config, resource):
        # type: (Config, models.AutoGenIAMPolicy) -> None
        if isinstance(resource.document, models.Placeholder):
            policy = self._policy_gen.generate_policy(config)
            if models.RoleTraits.VPC_NEEDED in resource.traits:
                policy['Statement'].append(VPC_ATTACH_POLICY)
            resource.document = policy


class BuildStage(object):
    def __init__(self, steps):
        # type: (List[BaseDeployStep]) -> None
        self._steps = steps

    def execute(self, config, resources):
        # type: (Config, List[models.Model]) -> None
        for resource in resources:
            for step in self._steps:
                step.handle(config, resource)


class ResultsRecorder(object):
    def __init__(self, osutils):
        # type: (OSUtils) -> None
        self._osutils = osutils

    def record_results(self, results, chalice_stage_name, project_dir):
        # type: (Any, str, str) -> None
        deployed_dir = self._osutils.joinpath(
            project_dir, '.chalice', 'deployed')
        deployed_filename = self._osutils.joinpath(
            deployed_dir, '%s.json' % chalice_stage_name)
        if not self._osutils.directory_exists(deployed_dir):
            self._osutils.makedirs(deployed_dir)
        serialized = serialize_to_json(results)
        self._osutils.set_file_contents(
            filename=deployed_filename,
            contents=serialized,
            binary=False
        )


class NoopResultsRecorder(ResultsRecorder):
    def record_results(self, results, chalice_stage_name, project_dir):
        # type: (Any, str, str) -> None
        return None


class DeploymentReporter(object):
    # We want the API URLs to be displayed last.
    _SORT_ORDER = {
        'rest_api': 100,
        'websocket_api': 100,
        'domain_name': 100
    }
    # The default is chosen to sort before the rest_api
    _DEFAULT_ORDERING = 50

    def __init__(self, ui):
        # type: (UI) -> None
        self._ui = ui

    def generate_report(self, deployed_values):
        # type: (Dict[str, Any]) -> str
        report = [
            'Resources deployed:',
        ]
        ordered = sorted(
            deployed_values['resources'],
            key=lambda x: self._SORT_ORDER.get(x['resource_type'],
                                               self._DEFAULT_ORDERING))
        for resource in ordered:
            getattr(self, '_report_%s' % resource['resource_type'],
                    self._default_report)(resource, report)
        report.append('')
        return '\n'.join(report)

    def _report_rest_api(self, resource, report):
        # type: (Dict[str, Any], List[str]) -> None
        report.append('  - Rest API URL: %s' % resource['rest_api_url'])

    def _report_websocket_api(self, resource, report):
        # type: (Dict[str, Any], List[str]) -> None
        report.append(
            '  - Websocket API URL: %s' % resource['websocket_api_url'])

    def _report_domain_name(self, resource, report):
        # type: (Dict[str, Any], List[str]) -> None
        report.append(
            '  - Custom domain name:\n'
            '      HostedZoneId: %s\n'
            '      AliasDomainName: %s' % (
                resource['hosted_zone_id'], resource['alias_domain_name']
            )
        )

    def _report_lambda_function(self, resource, report):
        # type: (Dict[str, Any], List[str]) -> None
        report.append('  - Lambda ARN: %s' % resource['lambda_arn'])

    def _report_lambda_layer(self, resource, report):
        # type: (Dict[str, Any], List[str]) -> None
        report.append('  - Lambda Layer ARN: %s' % (
            resource['layer_version_arn']))

    def _default_report(self, resource, report):
        # type: (Dict[str, Any], List[str]) -> None
        # The default behavior is to not report a resource.  This
        # cuts down on the output verbosity.
        pass

    def display_report(self, deployed_values):
        # type: (Dict[str, Any]) -> None
        report = self.generate_report(deployed_values)
        self._ui.write(report)



import sys
import warnings

from typing import Dict, List, Set, Iterator, Optional, Any  # noqa

from chalice import app  # noqa
from chalice.config import Config  # noqa
from chalice.constants import EXPERIMENTAL_ERROR_MSG
from chalice.constants import MIN_COMPRESSION_SIZE
from chalice.constants import MAX_COMPRESSION_SIZE
from chalice.compat import STRING_TYPES


class ExperimentalFeatureError(Exception):
    def __init__(self, features_missing_opt_in):
        # type: (Set[str]) -> None
        self.features_missing_opt_in = features_missing_opt_in
        msg = self._generate_msg(features_missing_opt_in)
        super(ExperimentalFeatureError, self).__init__(msg)

    def _generate_msg(self, missing_features):
        # type: (Set[str]) -> str
        opt_in_line = (
            'app.experimental_feature_flags.update([\n'
            '%s\n'
            '])\n' % ',\n'.join(["    '%s'" % feature
                                 for feature in missing_features]))
        return EXPERIMENTAL_ERROR_MSG % opt_in_line


def validate_configuration(config):
    # type: (Config) -> None
    """Validate app configuration.

    The purpose of this method is to provide a fail fast mechanism
    for anything we know is going to fail deployment.
    We can detect common error cases and provide the user with helpful
    error messages.

    """
    routes = config.chalice_app.routes
    validate_routes(routes)
    validate_route_content_types(routes, config.chalice_app.api.binary_types)
    validate_minimum_compression_size(config)
    _validate_manage_iam_role(config)
    validate_python_version(config)
    validate_unique_function_names(config)
    validate_feature_flags(config.chalice_app)
    validate_endpoint_type(config)
    validate_resource_policy(config)
    validate_sqs_configuration(config.chalice_app)
    validate_environment_variables_type(config)


def validate_resource_policy(config):
    # type: (Config) -> None
    if (config.api_gateway_endpoint_type != 'PRIVATE' and
            config.api_gateway_endpoint_vpce):
        raise ValueError(
            "config.api_gateway_endpoint_vpce should only be "
            "specified for PRIVATE api_gateway_endpoint_type")
    if config.api_gateway_endpoint_type != 'PRIVATE':
        return
    if config.api_gateway_policy_file and config.api_gateway_endpoint_vpce:
        raise ValueError(
            "Can only specify one of api_gateway_policy_file and "
            "api_gateway_endpoint_vpce")
    if config.api_gateway_policy_file:
        return
    if not config.api_gateway_endpoint_vpce:
        raise ValueError(
            ("Private Endpoints require api_gateway_policy_file or "
             "api_gateway_endpoint_vpce specified"))


def validate_endpoint_type(config):
    # type: (Config) -> None
    if not config.api_gateway_endpoint_type:
        return
    valid_types = ('EDGE', 'REGIONAL', 'PRIVATE')
    if config.api_gateway_endpoint_type not in valid_types:
        raise ValueError(
            "api gateway endpoint type must be one of %s" % (
                ", ".join(valid_types)))


def validate_feature_flags(chalice_app):
    # type: (app.Chalice) -> None
    missing_opt_in = set()
    # pylint: disable=protected-access
    for feature in chalice_app._features_used:
        if feature not in chalice_app.experimental_feature_flags:
            missing_opt_in.add(feature)
    if missing_opt_in:
        raise ExperimentalFeatureError(missing_opt_in)


def validate_routes(routes):
    # type: (Dict[str, Dict[str, app.RouteEntry]]) -> None
    # We're trying to validate any kind of route that will fail
    # when we send the request to API gateway.
    # We check for:
    #
    # * any routes that end with a trailing slash.
    for route_name, methods in routes.items():
        if not route_name:
            raise ValueError("Route cannot be the empty string")
        if route_name != '/' and route_name.endswith('/'):
            raise ValueError("Route cannot end with a trailing slash: %s"
                             % route_name)
        _validate_cors_for_route(route_name, methods)


def validate_python_version(config, actual_py_version=None):
    # type: (Config, Optional[str]) -> None
    """Validate configuration matches a specific python version.

    If the ``actual_py_version`` is not provided, it will default
    to the major/minor version of the currently running python
    interpreter.

    :param actual_py_version: The major/minor python version in
        the form "pythonX.Y", e.g "python2.7", "python3.6".

    """
    lambda_version = config.lambda_python_version
    if actual_py_version is None:
        actual_py_version = 'python%s.%s' % sys.version_info[:2]
    if actual_py_version != lambda_version:
        # We're not making this a hard error for now, but we may
        # turn this into a hard fail.
        warnings.warn("You are currently running %s, but the closest "
                      "supported version on AWS Lambda is %s\n"
                      "Please use %s, otherwise you may run into "
                      "deployment issues. " %
                      (actual_py_version, lambda_version, lambda_version),
                      stacklevel=2)


def validate_route_content_types(routes, binary_types):
    # type: (Dict[str, Dict[str, app.RouteEntry]], List[str]) -> None
    for methods in routes.values():
        for route_entry in methods.values():
            _validate_entry_content_type(route_entry, binary_types)


def _validate_entry_content_type(route_entry, binary_types):
    # type: (app.RouteEntry, List[str]) -> None
    binary, non_binary = [], []
    for content_type in route_entry.content_types:
        if content_type in binary_types:
            binary.append(content_type)
        else:
            non_binary.append(content_type)
    if binary and non_binary:
        # A routes content_types be homogeneous in their binary support.
        raise ValueError(
            'In view function "%s", the content_types %s support binary '
            'and %s do not. All content_types must be consistent in their '
            'binary support.' % (route_entry.view_name, binary, non_binary))


def _validate_cors_for_route(route_url, route_methods):
    # type: (str, Dict[str, app.RouteEntry]) -> None
    entries_with_cors = [
        entry for entry in route_methods.values() if entry.cors
    ]
    if entries_with_cors:
        # If the user has enabled CORS, they can't also have an OPTIONS
        # method because we'll create one for them.  API gateway will
        # raise an error about duplicate methods.
        if 'OPTIONS' in route_methods:
            raise ValueError(
                "Route entry cannot have both cors=True and "
                "methods=['OPTIONS', ...] configured.  When "
                "CORS is enabled, an OPTIONS method is automatically "
                "added for you.  Please remove 'OPTIONS' from the list of "
                "configured HTTP methods for: %s" % route_url)

        if not all(entries_with_cors[0].cors == entry.cors for entry in
                   entries_with_cors):
            raise ValueError(
                "Route may not have multiple differing CORS configurations. "
                "Please ensure all views for \"%s\" that have CORS configured "
                "have the same CORS configuration." % route_url
            )


def validate_minimum_compression_size(config):
    # type: (Config) -> None
    if config.minimum_compression_size is None:
        return
    if not isinstance(config.minimum_compression_size, int):
        raise ValueError("'minimum_compression_size' must be an int.")
    if config.minimum_compression_size < MIN_COMPRESSION_SIZE \
       or config.minimum_compression_size > MAX_COMPRESSION_SIZE:
        raise ValueError("'minimum_compression_size' must be equal to or "
                         "greater than %s and less than or equal to %s."
                         % (MIN_COMPRESSION_SIZE, MAX_COMPRESSION_SIZE))


def _validate_manage_iam_role(config):
    # type: (Config) -> None
    # We need to check if manage_iam_role is None because that's the value
    # it the user hasn't specified this value.
    # However, if the manage_iam_role value is not None, the user set it
    # to something, in which case we care if they set it to False.
    if not config.manage_iam_role:
        # If they don't want us to manage the role, they
        # have to specify an iam_role_arn.
        if not config.iam_role_arn:
            raise ValueError(
                "When 'manage_iam_role' is set to false, you "
                "must provide an 'iam_role_arn' in config.json."
            )


def validate_unique_function_names(config):
    # type: (Config) -> None
    names = set()   # type: Set[str]
    for name in _get_all_function_names(config.chalice_app):
        if name in names:
            raise ValueError("Duplicate function name detected: %s\n"
                             "Names must be unique across all lambda "
                             "functions in your Chalice app." % name)
        names.add(name)


def _get_all_function_names(chalice_app):
    # type: (app.Chalice) -> Iterator[str]
    for auth_handler in chalice_app.builtin_auth_handlers:
        yield auth_handler.name
    for event in chalice_app.event_sources:
        yield event.name
    for function in chalice_app.pure_lambda_functions:
        yield function.name


def validate_sqs_configuration(chalice_app):
    # type: (app.Chalice) -> None
    for event in chalice_app.event_sources:
        if not isinstance(event, app.SQSEventConfig):
            continue
        if not _is_valid_queue_name(event.queue, event.queue_arn):
            raise ValueError("The 'queue' parameter for the "
                             "'@app.on_sqs_message()' handler must be the "
                             "name of the queue, not the queue URL or the "
                             "queue ARN.  Invalid value: %s" % event.queue)


def _is_valid_queue_name(queue_name, queue_arn):
    # type: (Optional[str], Optional[str]) -> bool
    # The mutually exclusiveness is verified in the on_sqs_message decorator.
    if queue_name is not None and queue_name.startswith(('https:', 'arn:')):
        return False
    if queue_arn is not None and not queue_arn.startswith('arn:'):
        return False
    # We're not validating that the queue has only valid chars because SQS
    # won't let you create a queue with that name in the first place.  We just
    # want to detect the case where a user puts the queue URL/ARN instead of
    # the name for the queue_name.
    return True


def validate_environment_variables_type(config):
    # type: (Config) -> None
    _validate_environment_variables(config.environment_variables)
    for name in _get_all_function_names(config.chalice_app):
        _validate_environment_variables(
            config.scope(config.chalice_stage, name).environment_variables)


def _validate_environment_variables(environment_variables):
    # type: (Dict[str, Any]) -> None
    for key, value in environment_variables.items():
        if not isinstance(value, STRING_TYPES):
            raise ValueError("Environment variable values must be strings, "
                             "got 'type' %s for key '%s'" % (
                                 type(value).__name__, key))



import os

from chalice import Chalice

app = Chalice(app_name='{{app_name}}')
app.debug = True


# Set the value of APP_BUCKET_NAME in the .chalice/config.json file.
S3_BUCKET = os.environ.get('APP_BUCKET_NAME', '')


@app.on_s3_event(bucket=S3_BUCKET, events=['s3:ObjectCreated:*'])
def s3_handler(event):
    app.log.debug("Received event for bucket: %s, key: %s",
                  event.bucket, event.key)
